/***********************************************************************
 * Time-stamp: <Samstag, 12.05.2001, 17:18:20; edited by heiko>
 * 
 * Tina's random number generators
 *
 * lagged Fibonacci generator
 *
 * Copyright by Heiko Bauke, Magdeburg 2001 
 *
 * heiko.bauke@student.uni-magdeburg.de
 *
 * This is free software. You may use and distribute this software under
 * the terms of the GNU GENERAL PUBLIC LICENSE Version 2. See COPYING
 * for details.
 *
 ***********************************************************************/

#ifndef FIBONACCI_H
#define FIBONACCI_H

#include <iostream>
#include <strstream>
#include <new>
#include "trng.h"

class Fibonacci : public TRNG::RNG {
private:
  std::vector<long> r;
  long p, q, pointer;
  TRNG::ParkMiller R;
public:
  static const TRNG::RNG_type type=TRNG::user1_t;
  const char * name(void);
  void reset(void);
  void seed(long);
  long rand(void);
  RNG *copy(void);
//    void split(long, long);
//    void jump(long);
  void save_status(std::vector<long> &);
  void load_status(std::vector<long> &);
  Fibonacci(long, long);
  virtual ~Fibonacci() {};
};

inline const char * Fibonacci::name(void) {
  std::ostrstream out;
  out << "Fibonacci " << p;
  return((const char *)out.str());
}

inline void Fibonacci::reset(void) {
  seed(0l);
  max_val=0x7fffffffl;
  max_val2=max_val/2l;
}

inline void Fibonacci::seed(long s) {
  long i, j, h1, h2;
  h1=0x7fffffffl;
  h2=0x40000000l;
  R.seed(s);
  for (i=0; i<=p; ++i) {
//      if (i<q) {
//        r[i]=((R.vall()+1l) & h1) | h2;
//        h1>>=1;
//        h2>>=1;
//        if (h2==0) {
//      h1=0x7fffffffl;
//      h2=0x40000000l;
//        }
//      } else
//      r[i]=R.vall()+1l;
    r[i]=0l;
    for (j=0l; j<31; ++j) {
      r[i]<<=1;
      if (R.boolean())
        ++r[i];
    }
  }
  pointer=p-1l;
}

inline long Fibonacci::rand() {
  pointer=(pointer+1l)%p;
  return(r[pointer]=r[pointer]^r[(pointer+q)%p]);
}

TRNG::RNG *Fibonacci::copy(void) {
  std::vector<long> s;
  TRNG::RNG *R=static_cast<RNG *>(new Fibonacci(p, q));
  save_status(s);
  R->load_status(s);
  return(R);
}

void Fibonacci::save_status(std::vector<long> &s) {
  long i;
  s.resize(r.size()+2);
  s[0]=type;
  s[1]=q;
  for (i=0; i<p; ++i)
    s[i+2]=r[i];
}

void Fibonacci::load_status(std::vector<long> &s) {
  long i;
  if (s[0]!=type)
    TRNG::warn("Fibonacci::load_status wrong parameter");
  else {
    p=s.size()-2l;
    r.resize(p);
    q=s[1];
    for (i=0; i<p; ++i)
      r[i]=s[i+2];
  }
}

inline Fibonacci::Fibonacci(long p1, long q1) : r() {
  if (p1<0l || q1<0l || p1==q1)
    TRNG::die("bad arguments for Fibonacci::Fibonacci");
  if (p1>q1) {
    p=p1;
    q=q1;
  } else {
    p=q1;
    q=p1;
  }
  r.resize(p);
  reset();
}

#endif
