/***********************************************************************
 * Time-stamp: <Mittwoch, 09.05.2001, 14:31:36; edited by heiko>
 * 
 * Tina's random number generators
 *
 * Copyright by Heiko Bauke, Magdeburg 2001 
 *
 * testing random number generators by random walks
 *
 * heiko.bauke@student.uni-magdeburg.de
 *
 * This is free software. You may use and distribute this software under
 * the terms of the GNU GENERAL PUBLIC LICENSE Version 2. See COPYING
 * for details.
 *
 ***********************************************************************/

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <cmath>
#include <new>
#include <vector>
#include <set>
#include "trng.h"

using namespace TRNG;
using namespace std;


double regression(vector<double> &x, vector<double> &y, long n) {
  double t1, t2, t3, t4;
  long i;
  t1=0.0;
  t2=0.0;
  t3=0.0;
  t4=0.0;
  for (i=0l; i<n; ++i) {
    t1+=x[i]*y[i];
    t2+=x[i];
    t3+=y[i];
    t4+=x[i]*x[i];
  }
  return((t1-t2*t3/n)/(t4-t2*t2/n));
}


void multiwalk(RNG **r, long num_w, long runs) {
  long i, j, max, min;
  vector<long> pos;
  vector<double> x, y;
  x.resize(runs, 0.0);
  y.resize(runs, 0.0);
  pos.resize(num_w, 0l);
  for (j=0l; j<num_w; j++) 
    r[j]->split(num_w, j);
  min=0l;
  max=0l;
  for (j=0l; j<runs; ++j) {
    for (i=0l; i<num_w; ++i) {
      if (r[i]->boolean())
	++pos[i];
      else
	--pos[i];
      if (pos[i]<min)
	min=pos[i];
      if (pos[i]>max)
	max=pos[i];
    }
    x[j]=log(static_cast<double>(j+1l));
    y[j]=log(static_cast<double>(max-min));
    if (j%10l==9l)
      cout << j+1l << '\t' << regression(x, y, j) << endl;
  }
}


int main(void) {
  char c;
  long num_w, runs, i;
  RNG **r;
  
  cout << "number of walkers: ";
  cin >> num_w;
  r=new(RNG*[num_w]);
  cout << "choose a pseudo random number generator\n\n";
  cout << "  a) ParkMiller\n";
  cout << "  b) LMC2\n";
  cout << "  c) LMC3\n";
  cout << "  d) LMC4\n";
  cout << "  e) LCG64\n";
  cout << "  f) CLCG2\n";
  cout << "  g) CLCG3\n";
  cout << "  h) CLCG4\n";
  cout << "  i) RAND\n\n";
  cin >> c;
  switch (c) {
      case 'a' : 
	for (i=0l; i<num_w; ++i)
	  r[i]=new(ParkMiller);
	break;
      case 'b' : 
	for (i=0l; i<num_w; ++i)
	  r[i]=new(LMC2); 
	break;
      case 'c' : 
	for (i=0l; i<num_w; ++i)
	  r[i]=new(LMC3); 
	break;
      case 'd' : 
	for (i=0l; i<num_w; ++i)
	  r[i]=new(LMC4); 
	break;
      case 'e' : 
	for (i=0l; i<num_w; ++i)
	  r[i]=new(LCG64); 
	break;
      case 'f' : 
	for (i=0l; i<num_w; ++i)
	  r[i]=new(CLCG2); 
	break;
      case 'g' : 
	for (i=0l; i<num_w; ++i)
	  r[i]=new(CLCG3); 
	break;
      case 'h' : 
	for (i=0l; i<num_w; ++i)
	  r[i]=new(CLCG4); 
	break;
      case 'i' : 
	for (i=0l; i<num_w; ++i)
	  r[i]=new(RAND); 
	break;
      default : return(EXIT_FAILURE);
  }
  cout << "number of runs: ";
  cin >> runs;
  multiwalk(r, num_w, runs);
  return(EXIT_SUCCESS);
}












