// *************************************************************
// 
// eine Beispielapplikation 
//
// Monte-Carlo-pi-Berechnung
// 
// *************************************************************


#include <cstdlib>
#include <cmath>
#include <iostream>
#include "trng.h"
#include "mpi++.h"

using namespace TRNG;
using namespace std;

int main(int argc, char *argv[]) {

  // Gesamtzahl der Punkte im Quadrat
  const long all_samples=100000l;

  // Variablen
  long i, in, num_samples;
  double x, y, pi, pi_mean, pi_sigma;

  // ein Pseudozufallszahlengeneratorobjekt
  EINV r;
  
  // MPI initialisieren
  MPI::Init(argc, argv);

  // eigenen Rang und Geasamtzahl aller Prozesse bestimmen
  int size=MPI::COMM_WORLD.Get_size();
  int rank=MPI::COMM_WORLD.Get_rank();
  
  // Pseudozufallszahlenfolge auf die Prozesse aufteilen
  r.split(size, rank);
  
  // bisher keine Punke im Halbkreis
  in=0l;

  // Anzahl der je Prozess zu ermittelnden Punkte
  num_samples=all_samples/size+1l;
  for (i=0l; i<num_samples; ++i) {
    x=r.uniform();
    y=r.uniform();
    // liegt Punkt im Halbkreis
    if (x*x+y*y<=1.0)
      // ja? inkrementiere in
      ++in;
  }

  // pi berechnen
  pi=4.0*static_cast<double>(in)/static_cast<double>(num_samples);

  // eigener Rang 0?
  if (rank==0) {
    // ja, dann Ergebnisse der anderen Prozesse einsmmeln
    // Feld anlegen
    vector<double> pi_array(size);
    // eigenes Ergebnis ausgeben
    cout << "rank " << 0 << '\t' << pi << endl;
    pi_array[0]=pi;
    pi_mean=pi;
    // einsammeln und ausgeben
    for (i=1l; i<size; ++i) {
      MPI::COMM_WORLD.Recv(&pi, 1, MPI::DOUBLE, i, 0);
      cout << "rank " << i << '\t' << pi << endl;
      pi_array[i]=pi;
      pi_mean+=pi;
    }
    // Mittelwert berechnen und ausgeben
    pi_mean/=static_cast<double>(size);
    cout << "------------" << endl;
    cout << "mean\t" << pi_mean << endl;
    // mehr als ein Prozess?
    if (size>1) {
      // dann Standardabweichung berechnen und ausgeben
      pi_sigma=0.0;
      for (i=0l; i<size; ++i)
	pi_sigma+=(pi_array[i]-pi_mean)*(pi_array[i]-pi_mean);
      pi_sigma/=(static_cast<double>(size)-1.0)
                *static_cast<double>(size);
      pi_sigma=sqrt(pi_sigma);
      cout << "sigma\t" << pi_sigma << endl; 
    }
  } else
    // sonst Ergebnis versenden
    MPI::COMM_WORLD.Send(&pi, 1, MPI::DOUBLE, 0, 0);

  // MPI beenden
  MPI::Finalize();
  
  return(EXIT_SUCCESS);
}
